// # Inject Decorator
// Decorator for adding dependencies to a provider
//
// ## Usage
// Inject string-based dependencies, decorated classes, or classes directly as services
// ```js
// import {Inject} from 'ng-forard';
//
// class SomeService{ }
//
// @Inject('$q', SomeService)
// class AnotherService{ }
// ```
//
// Import the appWriter for reading and writing metadata
'use strict';

Object.defineProperty(exports, '__esModule', {
	value: true
});

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i]; return arr2; } else { return Array.from(arr); } }

var _writers = require('../writers');

var _utilGetInjectableName = require('../util/get-injectable-name');

var _decoratorsProviders = require('../decorators/Providers');

var _classesOpaqueToken = require('../classes/opaque-token');

// ## @Inject
// Takes an array of injects
var Inject = function Inject() {
	for (var _len = arguments.length, injects = Array(_len), _key = 0; _key < _len; _key++) {
		injects[_key] = arguments[_key];
	}

	return function (t) {
		var notStringBased = function notStringBased(inj) {
			return typeof inj !== 'string' && !(inj instanceof _classesOpaqueToken.OpaqueToken);
		};
		var ensureInjectable = function ensureInjectable(inj) {
			if (!_writers.providerWriter.get('name', inj) || !_writers.providerWriter.get('type', inj)) {
				throw new Error('Processing "' + t.name + '" @Inject parameter: "' + (inj.name || inj.toString()) + '" is not a valid injectable.\n\t\t\tPlease ensure ' + (inj.name || inj.toString()) + ' is injectable. Valid examples can be:\n\t\t\t - a string representing an ng1 provider, e.g. \'$q\'\n\t\t\t - an @Injectable ng-forward class\n\t\t\t - a Provider, e.g. provide(SOME_CONFIG, {asValue: 100})');
			}
			return inj;
		};
		// At the end of the day, Angular 1's DI requires the injection array to be
		// an array of strings. Map over the injects to get the string provider name for
		// each injectable

		var providers = injects.filter(notStringBased).map(ensureInjectable);

		_decoratorsProviders.Providers.apply(undefined, _toConsumableArray(providers))(t);

		var dependencies = injects.map(_utilGetInjectableName.getInjectableName).filter(function (n) {
			return n !== undefined;
		});

		// If there is already an $inject array, assume that it was set by a parent class.
		// The resultant $inject array should be a concat of local dependencies and parent
		// injects.
		// ```js
		// @Inject('$q', '$http', '$interval')
		// class Parent{ ... }
		//
		// @Inject('$timeout')
		// class Child extends Parent{
		// 	constructor($timeout, ...parentDependencies){
		// 		super(...parentDependencies);
		// 	}
		// }
		// ```
		if (_writers.appWriter.has('$inject', t)) {
			var parentInjects = _writers.appWriter.get('$inject', t);
			_writers.appWriter.set('$inject', [].concat(_toConsumableArray(dependencies), _toConsumableArray(parentInjects)), t);
		}
		// Otherwise just use the dependencies array as the $inject array.
		else {
				_writers.appWriter.set('$inject', dependencies, t);
			}
	};
};
exports.Inject = Inject;