// # Component Decorator
// Provides a robust component decorator that attempts to polyfill many Angular 2
// features while still providing an excellent, testable Angular 1 component strategy
//
// ## Usage
// ```js
// import {Component, EventEmitter, Inject} from 'ng-forward';
// import {Messenger} from '../messenger';
//
// @Component({
// 	selector: 'send-message',
// 	outputs: ['sent'],
// 	inputs: ['messageSubject: subject'],
// 	bind: [Messenger]
// 	template: `
// 		<textarea ng-model="sendMessage.body"></textarea>
// 		<button on-click="sendMessage.send()"></textarea>
// 	`
// })
// @Inject(Messenger)
// export class SendMessage{
// 	this.sent = new EventEmitter();
// 	constructor(Messenger){
// 		this.Messenger = Messenger;
// 		this.message = '';
// 	}
// 	async send(){
// 		let message = await this.Messenger.create(this.subject, this.message);
// 		this.sent.next(message);
// 	}
// }
//```
// In your HTML:
// ```html
// <send-message subject="Hello, World!" on-sent="sent($event)"></send-message>
// ```
// ## Setup
// `parseSelector` takes some simple CSS selector and returns a camelCased version
// of the selector as well as the type of selector it was (element, attribute, or
// CSS class).
'use strict';

Object.defineProperty(exports, '__esModule', {
	value: true
});

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i]; return arr2; } else { return Array.from(arr); } }

var _utilParseSelector = require('../../util/parse-selector');

var _utilParseSelector2 = _interopRequireDefault(_utilParseSelector);

// `providerWriter` sets up provider information, `componentWriter` writes the DDO,
// and `appWriter` sets up app traversal/bootstrapping information.

var _writers = require('../../writers');

// Takes the information from `config.providers` and turns it into the actual metadata
// needed during app traversal

var _Providers = require('../Providers');

// Provider parser will need to be registered with Module

var _module2 = require('../../module');

var _module3 = _interopRequireDefault(_module2);

var _utilParsePropertyMap = require('../../util/parse-property-map');

var _utilParsePropertyMap2 = _interopRequireDefault(_utilParsePropertyMap);

var _utilEvents = require('../../util/events');

var _utilEvents2 = _interopRequireDefault(_utilEvents);

var _utilDirectiveController = require('../../util/directive-controller');

var _utilDirectiveController2 = _interopRequireDefault(_utilDirectiveController);

var _utilInputsBuilder = require('../../util/inputs-builder');

var _extend = require('extend');

var _extend2 = _interopRequireDefault(_extend);

var _componentView = require('../component/view');

var TYPE = 'component';

// ## Decorator Definition
var Component = function Component(componentConfig) {
	return function (t) {
		// The only required config is a selector. If one wasn't passed, throw immediately
		if (!componentConfig.selector) {
			throw new Error('Component Decorator Error in "' + t.name + '": Component selector must be provided');
		}

		var DEFAULT_CONFIG = {
			inputs: [],
			providers: [],
			directives: [],
			outputs: []
		};

		var config = (0, _extend2['default'])({}, DEFAULT_CONFIG, componentConfig || {});

		// Grab the provider name and selector type by parsing the selector

		var _parseSelector = (0, _utilParseSelector2['default'])(config.selector);

		var name = _parseSelector.name;
		var restrict = _parseSelector.type;

		// Setup provider information using the parsed selector
		_writers.providerWriter.set('name', name, t);
		_writers.providerWriter.set('type', TYPE, t);

		// The appWriter needs the raw selector. This lets it bootstrap the root component
		_writers.appWriter.set('selector', config.selector, t);

		// Grab the providers from the config object, parse them, and write the metadata
		// to the target.
		_Providers.Providers.apply(undefined, _toConsumableArray(config.providers))(t);

		// Restrict type must be 'element'
		_writers.componentWriter.set('restrict', restrict, t);

		// Components should always create an isolate scope
		_writers.componentWriter.set('scope', {}, t);

		// Inputs should always be bound to the controller instance, not
		// to the scope
		_writers.componentWriter.set('bindToController', true, t);

		// Must perform some basic shape checking on the config object
		['inputs', 'providers', 'directives', 'outputs'].forEach(function (property) {
			if (config[property] !== undefined && !Array.isArray(config[property])) {
				throw new TypeError('Component Decorator Error in "' + t.name + '": Component ' + property + ' must be an array');
			}
		});

		// Check for Angular 2 style inputs
		var inputMap = (0, _utilParsePropertyMap2['default'])(config.inputs);
		var previousInputMap = _writers.componentWriter.get('inputMap', t) || {};
		_writers.componentWriter.set('inputMap', (0, _extend2['default'])({}, previousInputMap, inputMap), t);

		// outputs
		if (config.outputs.length > 0) {
			var outputMap = (0, _utilParsePropertyMap2['default'])(config.outputs) || {};
			_writers.componentWriter.set('outputMap', outputMap, t);
			for (var key in outputMap) {
				_utilEvents2['default'].add(outputMap[key]);
			}
		}

		// Allow for renaming the controllerAs
		if (config.controllerAs) {
			_writers.componentWriter.set('controllerAs', config.controllerAs, t);
		} else {
			// ControllerAs is the parsed selector. For example, `app` becomes `app` and
			// `send-message` becomes `sendMessage`
			_writers.componentWriter.set('controllerAs', name, t);
		}

		// Set a link function
		if (t.link) {
			_writers.componentWriter.set('link', t.link, t);
		}

		// Set a compile function
		if (t.compile) {
			_writers.componentWriter.set('compile', t.compile, t);
		}

		// Extract the view information
		// (View was merged with Component. See https://github.com/angular/angular/pull/4566)
		(0, _componentView.View)(config)(t);
	};
};

exports.Component = Component;
// ## Component Provider Parser
_module3['default'].addProvider(TYPE, function (target, name, injects, ngModule) {
	// First create an empty object to contain the directive definition object
	var ddo = {};

	_writers.componentWriter.forEach(function (val, key) {
		// Loop through the key/val pairs of metadata and assign it to the DDO
		ddo[key] = val;
	}, target);

	// Get the inputs bindings ahead of time
	var bindProp = angular.version.minor >= 4 ? 'bindToController' : 'scope';
	ddo[bindProp] = (0, _utilInputsBuilder.inputsMap)(ddo.inputMap);

	checkComponentConfig();

	// Component controllers must be created from a factory. Checkout out
	// util/directive-controller.js for more information about what's going on here
	ddo.controller = ['$scope', '$element', '$attrs', '$transclude', '$injector', function ($scope, $element, $attrs, $transclude, $injector) {
		var instance = (0, _utilDirectiveController2['default'])(this, injects, target, ddo, $injector, {
			$scope: $scope,
			$element: $element,
			$attrs: $attrs,
			$transclude: $transclude
		});

		return instance;
	}];

	// Finally add the component to the raw module
	ngModule.directive(name, function () {
		return ddo;
	});

	function createConfigErrorMessage(message) {
		return 'Processing "' + target.name + '" in "' + ngModule.name + '": ' + message;
	}

	function checkComponentConfig() {
		// If the selector type was not an element, throw an error. Components can only
		// be elements in Angular 2, so we want to enforce that strictly here.
		if (ddo.restrict !== 'E') {
			throw new Error(createConfigErrorMessage('@Component selectors can only be elements. ' + 'Perhaps you meant to use @Directive?'));
		}

		if (!ddo.templateUrl && !ddo.template) {
			throw new Error(createConfigErrorMessage('Components must have a template or templateUrl via the @Component or @View decorators'));
		}
	}
});