/* global it, describe */
'use strict';

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

require('../tests/frameworks');

var _inputsBuilder = require('./inputs-builder');

var _decoratorsProvidersComponent = require('../decorators/providers/component');

var _decoratorsComponentView = require('../decorators/component/view');

var _decoratorsInject = require('../decorators/inject');

var _bundle = require('../bundle');

var _bundle2 = _interopRequireDefault(_bundle);

var _bootstrap = require('../bootstrap');

var _bootstrap2 = _interopRequireDefault(_bootstrap);

var _testsAngular = require('../tests/angular');

describe('inputs-builder', function () {

  describe('inputsMap', function () {
    it('should return map of three isolate scope properties per input', function () {
      var inputs = {
        foo: 'foo'
      };

      var definition = (0, _inputsBuilder.inputsMap)(inputs);
      expect(definition).to.eql({
        '_bind_string_foo': '@foo',
        '[foo]': '&',
        '[(foo)]': '=?'
      });
    });

    it('should return map of inputs with local vs. public name', function () {
      var inputs = {
        fooLocal: 'fooPublic'
      };

      var definition = (0, _inputsBuilder.inputsMap)(inputs);
      expect(definition).to.eql({
        '_bind_string_fooLocal': '@fooPublic',
        '[fooPublic]': '&',
        '[(fooPublic)]': '=?'
      });
    });

    it('should return empty map if there are no inputs', function () {
      var inputs = {};

      var definition = (0, _inputsBuilder.inputsMap)(inputs);
      expect(definition).to.eql({});
    });
  });

  describe('inputsBuilder', function () {
    var controller = undefined;

    beforeEach(function () {
      controller = {};
    });

    describe('with localKey same as publicKey', function () {

      beforeEach(function () {
        (0, _inputsBuilder.inputsBuilder)(controller, 'foo', 'foo');
      });

      it('should add one visible and several hidden properties to controller', function () {
        var visibleProps = [];
        for (var key in controller) {
          visibleProps.push(key);
        }

        expect(controller).to.have.property('_bind_string_foo');
        expect(controller).to.have.property('[foo]');
        expect(controller).to.have.property('[(foo)]');
        expect(visibleProps).to.have.length(1);
        expect(visibleProps[0]).to.equal('foo');
      });

      it('should only be able to read from string input', function () {
        controller['[foo]'] = sinon.stub();
        // simulate angular setting value with hidden property
        controller['_bind_string_foo'] = 'bar';

        expect(controller.foo).to.equal('bar');
        controller.foo = 'quux'; // this has no effect;
        expect(controller.foo).to.equal('bar');
      });

      it('should only be able to read from one-way input', function () {
        // simulate angular one way fn binding, special for one-way only
        controller['[foo]'] = sinon.stub().returns('bar');

        expect(controller.foo).to.equal('bar');
        controller.foo = 'quux'; // this has no effect;
        expect(controller.foo).to.equal('bar');
      });

      it('should be able to read and write a two-way input', function () {
        // simulate angular setting value with hidden property
        controller['[foo]'] = sinon.stub();
        // simulate angular one way fn binding, special for one-way only
        controller['[(foo)]'] = 'bar';

        expect(controller.foo).to.equal('bar');
        controller.foo = 'quux';
        expect(controller.foo).to.equal('quux');
        expect(controller['[(foo)]']).to.equal('quux');
      });

      it('should allow writing to a two-way input that is initialized to a falsy defined value', function () {
        // simulate angular setting value with hidden property
        controller['[foo]'] = sinon.stub();
        // simulate angular one way fn binding, special for one-way only
        controller['[(foo)]'] = '';

        expect(controller.foo).to.equal('');
        controller.foo = 'quux';
        expect(controller.foo).to.equal('quux');
        expect(controller['[(foo)]']).to.equal('quux');
      });

      it('should not allow using more than one binding type', function () {
        controller['[foo]'] = sinon.stub().returns('bar');
        expect(controller['[(foo)]']).to['throw'];
      });
    });

    describe('with localKey different than publicKey', function () {
      beforeEach(function () {
        (0, _inputsBuilder.inputsBuilder)(controller, 'fooLocal', 'fooPublic');
      });

      it('should add one visible and several hidden properties to controller', function () {
        var visibleProps = [];
        for (var key in controller) {
          visibleProps.push(key);
        }

        expect(controller).to.have.property('_bind_string_fooLocal');
        expect(controller).to.have.property('[fooPublic]');
        expect(controller).to.have.property('[(fooPublic)]');
        expect(visibleProps).to.have.length(1);
        expect(visibleProps[0]).to.equal('fooLocal');
      });

      it('should only be able to read from string input', function () {
        controller['[fooPublic]'] = sinon.stub();
        // simulate angular setting value with hidden property
        controller['_bind_string_fooLocal'] = 'bar';

        expect(controller.fooLocal).to.equal('bar');
        controller.fooLocal = 'quux'; // this has no effect;
        expect(controller.fooLocal).to.equal('bar');
      });

      it('should only be able to read from one-way input', function () {
        // simulate angular one way fn binding, special for one-way only
        controller['[fooPublic]'] = sinon.stub().returns('bar');

        expect(controller.fooLocal).to.equal('bar');
        controller.fooLocal = 'quux'; // this has no effect;
        expect(controller.fooLocal).to.equal('bar');
      });

      it('should be able to read and write a two-way input', function () {
        // simulate angular setting value with hidden property
        controller['[fooPublic]'] = sinon.stub();
        // simulate angular one way fn binding, special for one-way only
        controller['[(fooPublic)]'] = 'bar';

        expect(controller.fooLocal).to.equal('bar');
        controller.fooLocal = 'quux';
        expect(controller.fooLocal).to.equal('quux');
        expect(controller['[(fooPublic)]']).to.equal('quux');
      });

      it('should not allow using more than one binding type', function () {
        controller['[fooPublic]'] = sinon.stub().returns('bar');
        expect(controller['[(fooPublic)]']).to['throw'];
      });
    });
  });
});